"""
==========
LAISSEZ
==========
"""

import asyncio
import base64
import json
import logging
from fastmcp.server.http import StarletteWithLifespan
from eth_account import Account
from typing import List, cast
from starlette.responses import JSONResponse
from starlette.types import Message, Receive, Scope, Send
from x402.facilitator import FacilitatorClient
from x402.common import find_matching_payment_requirements, process_price_to_atomic_amount, x402_VERSION
from x402.types import PaymentPayload, PaymentRequirements, x402PaymentRequiredResponse, SupportedNetworks
from laissez.types import PaidTool


logger = logging.getLogger(__name__)


class RequestReplay:
    def __init__(self, receive: Receive, body: bytes):
        self._receive = receive
        self._body = body
        self._called = False
        self._lock = asyncio.Lock()

    async def __call__(self) -> Message:
        async with self._lock:
            if not self._called:
                self._called = True
                return {"type": "http.request", "body": self._body, "more_body": False}
        return await self._receive()


class LaissezMcpServerMiddleware:
    def __init__(self, app: StarletteWithLifespan, tools: List[PaidTool], wallet: Account):
        self.app = app
        self.tools = tools
        self.wallet = wallet
        self.facilitator = FacilitatorClient()

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return
        
        method = scope.get("method", "")
        path = scope.get("path", "")
        scheme = scope.get("scheme", "https")

        # Check if it's an MCP POST request
        if method != "POST" or not path.startswith("/mcp"):
            logger.debug("Non-MCP HTTP %s request to %s, passing through", method, path)
            await self.app(scope, receive, send)
            return
        
        # Parse the MCP request body
        body = b''
        more_body = True
        while more_body:
            message = await receive()
            if message['type'] == 'http.request':
                body += message.get('body', b'')
                more_body = message.get('more_body', False)
        
        receive_replay = RequestReplay(receive, body)

        # Parse the tool action
        tool_action = None
        if body:
            try:
                mcp_request = json.loads(body)
                tool_action = mcp_request.get("method")
            except json.JSONDecodeError:
                logger.warning("Failed to parse MCP request body: %s", body)
                
        if tool_action != "tools/call":
            logger.debug("MCP method %s does not require payment, passing through", tool_action)
            await self.app(scope, receive_replay, send)
            return

        # Parse the tool name
        tool_parameters = mcp_request.get("params", {})
        tool_name = tool_parameters.get("name")

        if not tool_name:
            logger.warning("MCP request missing tool name, passing through")
            await self.app(scope, receive_replay, send)
            return

        logger.debug("MCP request for tool %s", tool_name)

        # Check if the tool is in the payment config
        tool_names = [tool.name for tool in self.tools]
        if tool_name not in tool_names:
            logger.info("Tool %s not configured for payments, passing through", tool_name)
            await self.app(scope, receive_replay, send)
            return

        # Build the payment requirements
        headers = { k.decode().lower(): v.decode() for k, v in scope.get("headers", []) }
        tool = self.tools[tool_names.index(tool_name)]
        price = str(tool.price)
        network = tool.network

        try:
            max_amount, asset_address, eip712 = process_price_to_atomic_amount(price, network)
        except ValueError as e:
            error_response = JSONResponse(content={"error": f"Invalid price configuration: {str(e)}"}, status_code=500)
            await error_response(scope, receive_replay, send)
            return

        host = headers.get("host", "")
        resource_url = f"{scheme}://{host}{path}"
        
        payment_requirements = [
            PaymentRequirements(
                scheme="exact",
                network=cast(SupportedNetworks, network),
                asset=asset_address,
                max_amount_required=max_amount,
                resource=resource_url,
                description=tool.description,
                mime_type="application/json",
                pay_to=self.wallet.address,
                max_timeout_seconds=60,
                extra=eip712,
            )
        ]

        # Check if payment is required
        if 'x-payment' not in headers:
            logger.info("Payment required for tool %s", tool_name)

            response_data = x402PaymentRequiredResponse(
                x402_version=x402_VERSION,
                accepts=payment_requirements,
                error=f"Payment required for {tool_name}",
            ).model_dump(by_alias=True)

            error_response = JSONResponse(content=response_data, status_code=402)
            await error_response(scope, receive_replay, send)
            return

        # Decode the X-Payment header
        try:
            payment_dictionary = json.loads(base64.b64decode(headers['x-payment']))
            payment = PaymentPayload(**payment_dictionary)
        except Exception as e:
            logger.warning("Failed to decode X-Payment header: %s", e)
            error_response = JSONResponse(content={"error": f"Invalid payment header format: {str(e)}"}, status_code=400)
            await error_response(scope, receive_replay, send)
            return
            
        selected_payment_requirements = find_matching_payment_requirements(
            payment_requirements=payment_requirements,
            payment=payment
        )


        if not selected_payment_requirements:
            logger.warning("No matching payment requirements found for tool %s", tool_name)
            error_response = JSONResponse(content={"error": f"No matching payment requirements found"}, status_code=400)
            await error_response(scope, receive_replay, send)
            return

        # TODO
        # THIS IS FINE FOR BASE-SEPOLIA
        # BUT WILL NEED TO BE CALL TO LAISSEZ BACKEND FOR MAINNET
        verify_response = await self.facilitator.verify(payment, selected_payment_requirements)

        if not verify_response.is_valid:
            logger.warning("Payment verification failed for tool %s: %s", tool_name, verify_response.invalid_reason)
            error_response = JSONResponse(content={"error": f"Payment verification failed: {verify_response.invalid_reason}"}, status_code=402)
            await error_response(scope, receive_replay, send)
            return


        logger.info("Payment verified for tool %s", tool_name)

        async def send_wrapper(message: Message):
            if message['type'] == 'http.response.start':
                status_code = message['status']
                if 200 <= status_code < 300:
                    try:
                        settle_response = None
                        reason = 'unknown_settle_error'
                        for attempt in range(3):
                            try:
                                settle_response = await self.facilitator.settle(
                                    payment, selected_payment_requirements
                                )
                            except Exception as settle_error:
                                reason = f"settlement_exception:{settle_error}"
                                logger.exception(
                                    "Payment settlement attempt %s raised an exception for tool %s",
                                    attempt + 1,
                                    tool_name,
                                )
                                settle_response = None
                            else:
                                if settle_response and settle_response.success:
                                    break
                                reason = (
                                    settle_response.error_reason
                                    if settle_response
                                    else 'unknown_settle_error'
                                )
                                if attempt < 2:
                                    logger.warning(
                                        "Payment settlement attempt %s failed for tool %s: %s",
                                        attempt + 1,
                                        tool_name,
                                        reason,
                                    )
                            if settle_response and settle_response.success:
                                break
                            if attempt < 2:
                                await asyncio.sleep(0.5 * (attempt + 1))

                        if settle_response and settle_response.success:
                            settlement_payload = settle_response.model_dump_json(by_alias=True)
                            settlement_header = base64.b64encode(
                                settlement_payload.encode('utf-8')
                            ).decode('utf-8')
                            message['headers'].append(
                                (b'X-Payment-Response', settlement_header.encode('utf-8'))
                            )
                            message['headers'].append(
                                (b'Access-Control-Expose-Headers', b'X-Payment-Response')
                            )
                            logger.info("Payment settled for tool %s", tool_name)
                        else:
                            message['headers'].append(
                                (b'X-Payment-Error', reason.encode('utf-8'))
                            )
                            logger.warning("Payment settlement failed for tool %s: %s", tool_name, reason)
                    except Exception as e:
                        error_message = f"settlement_exception:{e}"
                        message['headers'].append(
                            (b'X-Payment-Error', error_message.encode('utf-8'))
                        )
                        logger.exception("Payment settlement raised an exception for tool %s", tool_name)
            await send(message)

        await self.app(scope, receive_replay, send_wrapper)


"""
==========
LAISSEZ
==========
"""

import asyncio
import json
import logging
from typing import Callable, Optional, Awaitable
from httpx import AsyncClient, Response
from x402.clients.base import MissingRequestConfigError, PaymentError, decode_x_payment_response, x402Client
from x402.clients.httpx import HttpxHooks
from x402.types import x402PaymentRequiredResponse
from laissez.types import PaymentLog


logger = logging.getLogger(__name__)


class LaissezHttpxHooks(HttpxHooks):
    def __init__(
        self,
        client: x402Client,
        on_payment: Optional[Callable[[PaymentLog], Awaitable[None]]] = None,
        before_payment: Optional[
            Callable[[x402PaymentRequiredResponse], Awaitable[None]]
        ] = None,
    ):
        super().__init__(client=client)
        self.on_payment = on_payment
        self.before_payment = before_payment
        self.payment_creation_lock = asyncio.Lock()

    async def on_response(self, response: Response) -> Response:
        if response.status_code != 402:
            return response

        if response.request and response.request.extensions.get("x402_retry"):
            return response

        try:
            if not response.request:
                raise MissingRequestConfigError("Missing request configuration")

            await response.aread()
            data = response.json()
            payment_response = x402PaymentRequiredResponse(**data)

            if self.before_payment:
                try:
                    approved, reason = await self.before_payment(payment_response)
                except Exception as e:
                    logger.warning("Payment guard hook raised an exception: %s", e)
                    return response

                if not approved:

                    request_id = None
                    tool_name = None
                    if response.request and response.request.content:
                        try:
                            request_data = json.loads(response.request.content)
                            request_id = request_data.get("id")
                            tool_name = (
                                request_data.get("params", {}).get("name")
                                if isinstance(request_data, dict)
                                else None
                            )
                        except json.JSONDecodeError:
                            pass

                    tool_label = tool_name or "unknown tool"
                    logger.info("Payment guard rejected payment for %s: %s", tool_label, reason)
                    rejection_message = (
                        f"Laissez guard blocked payment for '{tool_label}': {reason}. "
                        "Please finish the task without using that paid tool."
                    )

                    structured_detail = {
                        "type": "payment_rejected",
                        "tool": tool_name,
                        "reason": reason,
                    }

                    result_content = {
                        "jsonrpc": "2.0",
                        "id": request_id,
                        "result": {
                            "content": [
                                {
                                    "type": "text",
                                    "text": rejection_message,
                                }
                            ],
                            "structuredContent": structured_detail,
                            "isError": False,
                        },
                    }

                    response.status_code = 200
                    response.headers['Content-Type'] = 'application/json'
                    response._content = str.encode(json.dumps(result_content))
                    return response

            async with self.payment_creation_lock:
                selected_requirements = self.client.select_payment_requirements(
                    payment_response.accepts
                )

                payment_header = self.client.create_payment_header(
                    selected_requirements, payment_response.x402_version
                )

                request = response.request
                request.headers["X-Payment"] = payment_header
                request.extensions["x402_retry"] = True

                async with AsyncClient(follow_redirects=True, timeout=30.0) as client:
                    retry_response = await client.send(request)

                if 200 <= retry_response.status_code < 300:

                    if 'x-payment-error' in retry_response.headers:
                        logger.warning(
                            "Server reported payment settlement failure: %s",
                            retry_response.headers['x-payment-error'],
                        )
                        return retry_response


                    payment_info = "with unknown details"
                    decoded_header = {}
                    
                    if 'x-payment-response' in retry_response.headers:
                        try:
                            decoded_header = decode_x_payment_response(retry_response.headers['x-payment-response'])
                            payment_info = f"\nTransaction: {decoded_header.get('transaction')}\n\tPayer: {decoded_header.get('payer')}\n\tNetwork: {decoded_header.get('network')}"
                        except Exception as e:
                            payment_info = f"with settlement details that could not be decoded: {e}"
                            logger.debug("Failed to decode payment settlement details: %s", e)

                    logger.info("Payment authorized: %s", payment_info)

                    # post to laissez db
                    payment_log = PaymentLog(
                        transaction_hash=decoded_header.get("transaction", "unknown"),
                        network=selected_requirements.network,
                        description=selected_requirements.description,
                        paid_to=selected_requirements.pay_to,
                        paid_by=decoded_header.get("payer", self.client.account.address),
                        amount=selected_requirements.max_amount_required,
                        unit=selected_requirements.extra.get("name", "unknown"),
                    )

                    if self.on_payment:
                        try:
                            await self.on_payment(payment_log)
                        except Exception as e:
                            logger.warning("Failed to forward payment log: %s", e)
                            pass

                    response.status_code = retry_response.status_code
                    response.headers.clear()
                    response.headers.update(retry_response.headers)
                    response._content = await retry_response.aread()
                    response.request = retry_response.request

                    logger.debug("Payment header applied and request retried successfully")
                    return response
                else:
                    await retry_response.aread()
                    return retry_response

        except PaymentError as e:
            logger.error("Payment error while handling response: %s", e)
            return response
        except Exception as e:
            logger.exception("Unexpected error during payment handling")
            return response


"""
==========
LAISSEZ
==========

EXAMPLE: MCP Client with Laissez

This example shows how to connect to a paid MCP server using Laissez. 
All payments will be handled automatically and logged to your Laissez dashboard.

Pre-requisites:
* Ensure that you are running the paid MCP server from `examples/mcp/paid-mcp-server.py`.
    * You can do so from the root of the project by running `uv run examples/mcp/paid-mcp-server.py`.
* Update `BASE_URL` in this file to point to the URL of this example paid MCP server
* Ensure OPENAI_API_KEY is set in your environment

To run:
* Run this file from the root of the project by running `uv run examples/mcp/mcp-client.py`.
"""

import asyncio
from pydantic_ai import Agent
from laissez.mcp import LaissezMCPConnector
from dotenv import load_dotenv

load_dotenv()

LOCAL = False
BASE_URL = 'http://127.0.0.1:8000' if LOCAL else 'https://dice-roll.laissez.xyz'

async def main():

    server = LaissezMCPConnector(url=f'{BASE_URL}/mcp')

    agent = Agent(
        model='openai:gpt-5-mini',
        instructions='Show your working out',
        toolsets=[server]
    )

    result = await agent.run("Roll a dice twice then multiply the two results and return the final result")
    print(result.output)


if __name__ == '__main__':
    asyncio.run(main())


import os
import sys
import secrets
from eth_account import Account
from typing import Optional
import httpx
from x402.common import process_price_to_atomic_amount


def check_and_create_wallet(wallet: Optional[Account] = None) -> Account:
    if wallet is None:
        wallet_pk = os.getenv("WALLET_PRIVATE_KEY")
        if wallet_pk:
            wallet = Account.from_key(wallet_pk)
            print(f"[Laissez] Loaded wallet from .env file")
            print(f"[Laissez] Wallet address: {wallet.address}")
            print_usdc_balance_base_sepolia(wallet)
        else:
            wallet_creation_prompt = """
No wallet found. 
If you have a wallet private key, please set WALLET_PRIVATE_KEY in the .env file and run the program again.
If you don't have a wallet private key, you can create one now.
Note: If you choose to create a wallet, Laissez will add the wallet private key to the .env file.

Would you like to create a wallet now? (Y/n) >>> """
            wallet_creation_response = input(wallet_creation_prompt)
            if wallet_creation_response.lower() == "y" or wallet_creation_response.lower() == "":
                random_string = secrets.token_urlsafe(32)
                wallet = Account.create(random_string)
                env_file = ".env"
                with open(env_file, "a", encoding="utf-8") as f:
                    f.write(f"\nWALLET_PRIVATE_KEY={wallet.key.hex()}\n")
                    print("\n[Laissez] Wallet created and saved to .env file")

                print("="*50)
                print(f"[Laissez] SAVE THIS FOR FUTURE USE")
                print(f"[Laissez] Wallet address: {wallet.address}")
                print(f"[Laissez] Wallet private key: {wallet.key.hex()}")
                print("="*50)
                print()
                print("[Laissez] IMPORTANT:")
                print("[Laissez] Copy the wallet address and add faucet funds to 'Base Sepolia' for testing at https://faucet.circle.com")
                print()
                print("="*50)
                print("[Laissez] Please run the program again to continue.")
                print()
                sys.exit(0)
            else:
                raise ValueError("No wallet found and user did not want to create one")
    return wallet



def get_laissez_api_key(provided_api_key: Optional[str] = None) -> str:
    """Fetch and validate the Laissez API key.

    Prefers the provided_api_key argument, otherwise reads LAISSEZ_API_KEY from the environment.
    Ensures the key is in the expected format (lsz- prefix).
    """
    api_key = provided_api_key or os.getenv("LAISSEZ_API_KEY")
    if not api_key:
        raise ValueError(
            "LAISSEZ_API_KEY not found in environment. Create one at https://app.laissez.xyz to track agent spending."
        )
    if not api_key.startswith("lsz-"):
        raise ValueError(
            "Invalid LAISSEZ_API_KEY format. Get your API key at https://app.laissez.xyz"
        )
    return api_key


def _encode_balance_of_call(address: str) -> str:
    selector = "70a08231"  # keccak("balanceOf(address)") first 4 bytes
    addr = address.lower().replace("0x", "")
    padded_addr = addr.rjust(64, "0")
    return f"0x{selector}{padded_addr}"


def _get_usdc_address_base_sepolia() -> str:
    # Reuse x402 mapping for USDC on Base Sepolia
    _, asset_address, _ = process_price_to_atomic_amount("0.001", "base-sepolia")
    return asset_address


def print_usdc_balance_base_sepolia(wallet: Account, rpc_url: Optional[str] = None) -> None:
    """Print the wallet's USDC balance on Base Sepolia and exit.

    If the balance is 0, prints the faucet message and exits.
    """
    rpc = rpc_url or os.getenv("BASE_SEPOLIA_RPC_URL") or "https://sepolia.base.org"
    usdc_address = _get_usdc_address_base_sepolia()
    call_data = _encode_balance_of_call(wallet.address)

    payload = {
        "jsonrpc": "2.0",
        "method": "eth_call",
        "params": [
            {"to": usdc_address, "data": call_data},
            "latest",
        ],
        "id": 1,
    }

    try:
        with httpx.Client(timeout=20.0) as client:
            resp = client.post(rpc, json=payload)
            resp.raise_for_status()
            data = resp.json()
    except Exception as e:
        print(f"[Laissez] Failed to get USDC balance on Base Sepolia: {e}")
        print(f"[Laissez] Check you have funds in your wallet at https://sepolia.basescan.org/address/{wallet.address}")

    result_hex = data.get("result")
    if not result_hex:
        print(f"[Laissez]Failed to get USDC balance on Base Sepolia: {data}")
        print(f"[Laissez] Check you have funds in your wallet at https://sepolia.basescan.org/address/{wallet.address}")

    raw_balance = int(result_hex, 16)

    if raw_balance == 0:
        print(f"[Laissez] USDC balance on Base Sepolia: 0.000000 USDC")
        print("[Laissez] Copy the wallet address and add faucet funds to 'Base Sepolia' for testing at https://faucet.circle.com")
        sys.exit(0)

    human = raw_balance / 10**6  # USDC has 6 decimals
    print(f"[Laissez] USDC balance on Base Sepolia: {human:.6f} USDC")
    print("[Laissez] You can add more funds to your wallet at https://faucet.circle.com")
    print("="*50)
    print()


def print_ascii_art() -> None:
    print("""
▐▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▀▌
▐                                                             ▌
▐                                                             ▌
▐                                                             ▌
▐     ██╗      █████╗ ██╗███████╗███████╗███████╗███████╗     ▌
▐     ██║     ██╔══██╗██║██╔════╝██╔════╝██╔════╝╚══███╔╝     ▌
▐     ██║     ███████║██║███████╗███████╗█████╗    ███╔╝      ▌
▐     ██║     ██╔══██║██║╚════██║╚════██║██╔══╝   ███╔╝       ▌
▐     ███████╗██║  ██║██║███████║███████║███████╗███████╗     ▌
▐     ╚══════╝╚═╝  ╚═╝╚═╝╚══════╝╚══════╝╚══════╝╚══════╝     ▌
▐                                                             ▌
▐                                                             ▌
▐                                                             ▌
▐▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▄▌
""")


"""
==========
LAISSEZ
==========

EXAMPLE: Paid MCP Server with Laissez

This example shows how to create a paid MCP server with Laissez.

Pre-requisites:
* Ensure that the names of the paid tools in the `paid_tools` list match the names of the tools in the MCP server

To run:
* Run this file from the root of the project by running `uv run examples/mcp/paid-mcp-server.py`.
"""

import random
import secrets
from fastmcp import FastMCP
from eth_account import Account
from laissez.types import PaidTool
from laissez.mcp import LaissezMCPProvider
from dotenv import load_dotenv

load_dotenv()


# STEP 1: CREATE YOUR MCP
mcp = FastMCP("Dice Roller")

@mcp.tool
def roll_die() -> int:
    return random.randint(1, 6)

@mcp.tool
def multiply(a: int, b: int) -> int:
    return a * b



# STEP 2: DEFINE YOUR PAID TOOLS
paid_tools = [
    PaidTool(name="roll_die", price=0.001, description="Roll a die and return the result"),
    PaidTool(name="multiply", price=0.005, description="Multiply two numbers and return the result"),
]



# STEP 3: CREATE OR BRING YOUR WALLET
random_string = secrets.token_urlsafe(32)
wallet = Account.create(random_string)



# STEP 4: RECEIVE PAYMENTS WITH LAISSEZ
app = LaissezMCPProvider(
    mcp=mcp,
    tools=paid_tools,
    wallet=wallet
)


if __name__ == "__main__":
    app.run()


"""
==========
LAISSEZ
==========
"""

from pydantic import BaseModel, Field
from typing import Optional, Literal

class PaymentLog(BaseModel):
    transaction_hash: str
    network: str
    description: str
    paid_to: str
    paid_by: str
    amount: str
    unit: str


class PaidTool(BaseModel):
    """
    Configure a paid tool for an MCP server.

    Args:
        name: The name of the tool
        price: The price of the tool in USDC. Defaults to 0.001.
        network: The network to use for the tool. Defaults to 'base-sepolia'.
        description: The description of the tool
    """
    name: str = Field(..., description="The name of the tool")
    price: float = Field(0.001, description="The price of the tool in USDC")
    network: Optional[Literal['base-sepolia', 'base']] = Field('base-sepolia', description="The network to use for the tool")
    description: str = Field(..., description="The description of the tool")


"""
==========
LAISSEZ
==========
"""

import asyncio
import logging
from typing import Optional, Tuple, List
from httpx import AsyncClient, Headers
from fastmcp import FastMCP
from pydantic_ai.mcp import MCPServerStreamableHTTP
from eth_account import Account

from x402.clients.base import x402Client
from x402.clients.httpx import x402HttpxClient as BaseX402HttpxClient
from x402.types import x402PaymentRequiredResponse

from laissez.types import PaymentLog, PaidTool
from laissez.common import check_and_create_wallet, get_laissez_api_key, print_ascii_art
from laissez.hooks import LaissezHttpxHooks
from laissez.middleware import LaissezMcpServerMiddleware

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger(__name__)


class LaissezMCPConnector(MCPServerStreamableHTTP):
    def __init__(
        self,
        url: str,
        api_key: Optional[str] = None,
        wallet: Optional[Account] = None,
    ):
        api_key = get_laissez_api_key(api_key)
        wallet = check_and_create_wallet(wallet)
        print_ascii_art()

        laissez_httpx_client = BaseX402HttpxClient(account=wallet, follow_redirects=True, timeout=30.0)
        x402_client = x402Client(account=wallet)
        
        self._backend = AsyncClient(
            base_url="https://app.laissez.xyz",
            timeout=30.0,
            headers=Headers({
                "Authorization": f"Bearer {api_key}",
                "Content-Type": "application/json",
                "User-Agent": "laissez-client-python/0.0.1a3",
            }),
        )

        async def _on_payment(payment_log: PaymentLog) -> None:
            try:
                payload = payment_log.model_dump()
                resp = await self._backend.post("/api/v0/log", json=payload)
                if resp.status_code >= 400:
                    logger.warning("Failed to post payment log: status=%s", resp.status_code)
                else:
                    logger.debug("Payment logged successfully")
            except Exception as e:
                logger.warning("Error posting payment log: %s", e)




        async def _before_payment(payment_response: x402PaymentRequiredResponse) -> Tuple[bool, str]:
            """
            Calls the laissez backend to check if payment should proceed.
            Raises an exception if payment is not approved.
            """
            try:
                payload = payment_response.model_dump()
                resp = await self._backend.post("/api/v0/guard", json=payload)
                if resp.status_code != 200:
                    return False, f"Payment guard check failed with status {resp.status_code}"

                data = resp.json()
                approved = data.get("approved", False)
                reason = data.get("reason", "No reason provided.")
                return approved, reason

            except Exception as e:
                logger.warning("Payment guard check error: %s", e)
                return False, str(e)


        hooks = LaissezHttpxHooks(
            client=x402_client, on_payment=_on_payment, before_payment=_before_payment
        )
        laissez_httpx_client.event_hooks = {
            "request": [hooks.on_request],
            "response": [hooks.on_response]
        }

        super().__init__(url=str(url), http_client=laissez_httpx_client)





class LaissezMCPProvider:
    def __init__(self, mcp: FastMCP, tools: List[PaidTool], wallet: Account):

        self.mcp = mcp
        self.wallet = check_and_create_wallet(wallet)

        mcp_tools = asyncio.run(mcp.get_tools())
        processed_tools = []
        for tool in tools:
            if tool.name not in mcp_tools:
                logger.info("Paid tool %s not available on MCP server; skipping", tool.name)
                continue
            else:
                processed_tools.append(tool)
        self.tools = processed_tools
        


    def run(self) -> None:
        app = self.mcp.http_app(
            transport='streamable-http',
            stateless_http=True
        )
        app.add_middleware(LaissezMcpServerMiddleware, tools=self.tools, wallet=self.wallet)
        print_ascii_art()
        import uvicorn
        uvicorn.run(app, host="127.0.0.1", port=8000, ws="none")